#include <stdlib.h>
#include <stdio.h>
#include <conio.h>
#include <string.h>

#define NTSC_CLOCK		1789773
#define NTSC_FRAMERATE	60
#define PAL_CLOCK		1662607
#define PAL_FRAMERATE	50

enum {
	FMT_VGM=0,
	FMT_TXT
};

enum {
	SYS_NTSC=0,
	SYS_PAL
};


struct {

	unsigned char *data;
	int size;

	int songs;
	int load_adr;
	int init_adr;
	int play_adr;

	unsigned char page[8];

} NSF;

struct {

	FILE* file;

	unsigned char ram[32768];
	unsigned char rom[1024*1024];
	int apu_regs[32];
	unsigned char page[8];

	bool change;
	bool done;
	int wait;
	int frame;

	unsigned char* data;
	int ptr;
	int size;

} CONV;

struct {

	int song_dur;
	int sub_song;
	bool cxx_stop;
	int system;
	int format;

} CONF;




static inline unsigned char mem_rd(int adr)
{
	if(adr<0x2000) return CONV.ram[adr&0x7ff];//RAM

	if(adr>=0x6000&&adr<0x8000) return CONV.ram[adr];//extra RAM

	if(adr>=0x8000) return CONV.rom[CONV.page[(adr-0x8000)>>12]*0x1000+(adr&0xfff)];//ROM page

	return 0;
}



static inline void mem_wr(int adr,unsigned char data)
{
	if(adr<0x2000)//RAM
	{
		CONV.ram[adr&0x7ff]=data;
		return;
	}

	if(adr>=0x5ff8&&adr<0x6000)	//NSF bankswitching
	{
		CONV.page[adr-0x5ff8]=data;
		return;
	}

	if(adr>=0x6000&&adr<0x8000)//extra RAM
	{
		CONV.ram[adr]=data;
		return;
	}

	if(adr<0x4017)//APU registers
	{
		if(adr==0x400a) CONV.apu_regs[0x0b]|=0x100;	//force triangle counter update

		if(CONV.apu_regs[adr-0x4000]!=data)
		{
			CONV.apu_regs[adr-0x4000]=(data&255)|0x100;

			CONV.change=true;
		}

		if(CONF.cxx_stop)	//catch the C00 effect
		{
			if(adr==0x4015&&!data) CONV.done=true;
		}
	}
}



#include "cpu2a03.h"



bool convert_vgm_begin(const char* filename,int subsong)
{
	CONV.file=fopen(filename,"wb");

	if(!CONV.file) return false;

	CONV.size=65536;
	CONV.data=(unsigned char*)malloc(CONV.size);
	CONV.ptr=256;

	memset(CONV.data,0,256);	//reserve for header

	return true;
}



void convert_vgm_frame(void)
{
	int id;

	for(id=0;id<32;++id)
	{
		if(CONV.apu_regs[id]&0x100)
		{
			CONV.apu_regs[id]&=0xff;

			CONV.data[CONV.ptr++]=0xb4;
			CONV.data[CONV.ptr++]=id;
			CONV.data[CONV.ptr++]=CONV.apu_regs[id];
		}
	}

	CONV.data[CONV.ptr++]=(CONF.system==SYS_NTSC)?0x62:0x63;

	if(CONV.ptr>=CONV.size-256)
	{
		CONV.size+=65536;
		CONV.data=(unsigned char*)realloc(CONV.data,CONV.size);
	}
}


void wr_dword(unsigned char* data,unsigned int n)
{
	data[0]=n&255;
	data[1]=(n>>8)&255;
	data[2]=(n>>16)&255;
	data[3]=(n>>24)&255;
}

void convert_vgm_end(void)
{
	int frame_samples;

	frame_samples=(CONF.system==SYS_NTSC)?44100/NTSC_FRAMERATE:44100/PAL_FRAMERATE;

	CONV.size=CONV.ptr;

	wr_dword(&CONV.data[0x00],0x206d6756);					//'Vgm ' signature
	wr_dword(&CONV.data[0x04],CONV.size-4);					//eof
	wr_dword(&CONV.data[0x08],0x00000161);					//version number
	wr_dword(&CONV.data[0x18],CONV.frame*frame_samples);	//samples
	wr_dword(&CONV.data[0x1c],CONV.frame*frame_samples);	//loop offset
	wr_dword(&CONV.data[0x20],0);							//loop samples
	wr_dword(&CONV.data[0x24],(CONF.system==SYS_NTSC)?NTSC_FRAMERATE:PAL_FRAMERATE);	//rate
	wr_dword(&CONV.data[0x34],0x100-0x34);					//vgm data relative offset
	wr_dword(&CONV.data[0x84],(CONF.system==SYS_NTSC)?NTSC_CLOCK:PAL_CLOCK);	//NES APU clock

	fwrite(CONV.data,CONV.size,1,CONV.file);
	fclose(CONV.file);

	free(CONV.data);
}



bool convert_txt_begin(const char* filename,int subsong)
{
	CONV.file=fopen(filename,"wt");

	if(!CONV.file) return false;

	if(CONF.system==SYS_NTSC)
	{
		fprintf(CONV.file,"system=NTSC\n");
		fprintf(CONV.file,"framerate=%i\n",NTSC_FRAMERATE);
		fprintf(CONV.file,"clock=%i\n",NTSC_CLOCK);
	}
	else
	{
		fprintf(CONV.file,"system=PAL\n");
		fprintf(CONV.file,"framerate=%i\n",PAL_FRAMERATE);
		fprintf(CONV.file,"clock=%i\n",PAL_CLOCK);
	}

	return true;
}



void convert_txt_frame(void)
{
	int id,cnt;

	if(!CONV.change)
	{
		++CONV.wait;
	}
	else
	{
		if(CONV.wait>0)
		{
			fprintf(CONV.file,"wait=%i\n",CONV.wait);

			CONV.wait=1;
		}

		cnt=0;

		for(id=0;id<32;++id)
		{
			if(CONV.apu_regs[id]&0x100)
			{
				CONV.apu_regs[id]&=0xff;

				if(cnt>0) fprintf(CONV.file,",");

				fprintf(CONV.file,"%2.2x=%2.2x",id,CONV.apu_regs[id]);

				++cnt;
			}
		}

		fprintf(CONV.file,"\n");
	}
}



void convert_txt_end(void)
{
	if(CONV.wait>0)
	{
		fprintf(CONV.file,"wait=%i\n",CONV.wait);
	}

	fclose(CONV.file);
}



bool convert_song(const char* filename,int subsong)
{
	int i,frame_dur;

	memset(&CONV,0,sizeof(CONV));
	memcpy(CONV.rom+(NSF.load_adr-0x8000),NSF.data+0x80,NSF.size-0x80);

	switch(CONF.format)
	{
	case FMT_VGM: if(!convert_vgm_begin(filename,subsong)) return false; break;
	case FMT_TXT: if(!convert_txt_begin(filename,subsong)) return false; break;
	}

	CONV.apu_regs[0x00]=0x3f;
	CONV.apu_regs[0x01]=0x08;
	CONV.apu_regs[0x04]=0x3f;
	CONV.apu_regs[0x05]=0x08;
	CONV.apu_regs[0x08]=0x80;
	CONV.apu_regs[0x0c]=0x3f;
	CONV.apu_regs[0x0f]=0x00;
	CONV.apu_regs[0x15]=0x1f;

	for(i=0;i<32;++i) CONV.apu_regs[i]|=0x100;

	for(i=0;i<8;++i) CONV.page[i]=NSF.page[i];

	CONV.done=false;
	CONV.change=false;
	CONV.wait=1;

	cpu_reset();

	CPU.A=subsong;
	CPU.X=(CONF.system==SYS_PAL)?1:0;
	CPU.PC.hl=NSF.init_adr;
	CPU.S=0xFC;				//reserve 3 bytes on the stack

	CONV.ram[0x1ff]=0x00;	//BRK instruction to cause jam
	CONV.ram[0x1fe]=0x01;	//return address 0x01FF-1
	CONV.ram[0x1fd]=0xfe;

	for(i=0;i<30000;++i)
	{
		cpu_tick();

		if(CPU.jam) break;
	}

	cpu_reset();

	frame_dur=CONF.song_dur*(CONF.system==SYS_NTSC?60:50);

	while(!CONV.done)
	{
		CPU.PC.hl=NSF.play_adr;
		CPU.jam=false;
		CPU.S=0xff;

		for(i=0;i<30000/4&&!CONV.done;++i)
		{
			cpu_tick();

			if(CPU.jam) break;
		}

		switch(CONF.format)
		{
		case FMT_VGM: convert_vgm_frame(); break;
		case FMT_TXT: convert_txt_frame(); break;
		}

		CONV.change=false;

		++CONV.frame;

		if(!CONF.cxx_stop)
		{
			--frame_dur;

			if(frame_dur<=0) break;
		}
	}

	switch(CONF.format)
	{
	case FMT_VGM: convert_vgm_end(); break;
	case FMT_TXT: convert_txt_end(); break;
	}

	return true;
}



bool nsf_load(char* filename)
{
	FILE *file;
	int i,page;
	bool bs;

	memset(&NSF,0,sizeof(NSF));

	file=fopen(filename,"rb");

	if(!file)
	{
		printf("Error: Can't open file\n");
		return false;
	}

	fseek(file,0,SEEK_END);
	NSF.size=ftell(file);
	fseek(file,0,SEEK_SET);
	NSF.data=(unsigned char*)malloc(NSF.size);
	fread(NSF.data,NSF.size,1,file);
	fclose(file);

	NSF.songs=NSF.data[0x06];

	NSF.load_adr=NSF.data[0x08]+(NSF.data[0x09]<<8);
	NSF.init_adr=NSF.data[0x0a]+(NSF.data[0x0b]<<8);
	NSF.play_adr=NSF.data[0x0c]+(NSF.data[0x0d]<<8);

	bs=false;

	for(i=0;i<8;++i)
	{
		page=NSF.data[0x70+i];
		NSF.page[i]=page;

		if(page) bs=true;
	}

	if(!bs)
	{
		for(i=0;i<8;++i) NSF.page[i]=i;
	}

	if(NSF.data[0x7b])
	{
		printf("Error: Expansion chips are not supported\n");
		free(NSF.data);
		return false;
	}

	return true;
}



int main(int argc,char *argv[])
{
	int i,id,song,option,song_from,song_to,in_arg,out_arg;
	char outname[1024],buf[16];

	if(argc<2)
	{
		printf("nsf2vgm v1.1 by Shiru (shiru@mail.ru) 11'19\n");
		printf("Usage: nsf2vgm [options] file.nsf [file.vgm]\n");
		printf("\t-sN process a particular sub song only\n");
		printf("\t-a processes all sub songs (default)\n");
		printf("\t-c stop at C00 command in FT (a $4015 change) instead of the fixed duration\n");
		printf("\t-dNNN is time in seconds before processing stops (180 seconds by default)\n");
		printf("\t-n force NTSC (default)\n");
		printf("\t-p force PAL\n");
		printf("\t-f set format, -fv (VGM, default), -ft (plain text), -fp (PSG)\n");

		return 1;
	}
	
	CONF.song_dur=3*60;
	CONF.sub_song=-1;
	CONF.cxx_stop=false;
	CONF.system  =SYS_NTSC;
	CONF.format  =FMT_VGM;

	id=1;
	in_arg=-1;
	out_arg=-1;

	while(id<argc)
	{
		if(argv[id][0]=='-')
		{
			option=argv[id][1];

			if(option>='A'&&option<='Z') option+=32;

			switch(option)
			{
			case 'a':
				CONF.sub_song=-1;
				break;

			case 's':
				CONF.sub_song=atoi(&argv[id][2]);
				break;

			case 'c':
				CONF.cxx_stop=true;
				break;

			case 'd':
				CONF.song_dur=atoi(&argv[id][2]);
				break;

			case 'n':
				CONF.system=SYS_NTSC;
				break;

			case 'p':
				CONF.system=SYS_PAL;
				break;

			case 'f':
				if(argv[id][2]=='v'||argv[id][2]=='V') CONF.format=FMT_VGM; else
				if(argv[id][2]=='t'||argv[id][2]=='T') CONF.format=FMT_TXT; else
				{
					printf("Error: Unknown output format\n");
					return 1;
				}
				break;
			}
		}
		else
		{
			if(in_arg<0) in_arg=id; else if(out_arg<0) out_arg=id;
		}

		++id;
	}

	if(in_arg<0) in_arg=1;
	if(out_arg<0) out_arg=in_arg;

	if(!nsf_load(argv[in_arg])) return 1;

	if(CONF.sub_song<0)
	{
		song_from=0;
		song_to  =NSF.songs;
	}
	else
	{
		song_from=CONF.sub_song;
		song_to  =CONF.sub_song+1;
	}

	for(song=song_from;song<song_to;++song)
	{
		strncpy(outname,argv[out_arg],sizeof(outname));
		for(i=strlen(outname)-1;i>=0;--i)
		{
			if(outname[i]=='.')
			{
				outname[i]=0;
				break;
			}
		}

		if(CONF.sub_song>=0)
		{
			sprintf(buf,"_%i",song+1);
			strncat(outname,buf,sizeof(outname));
		}

		switch(CONF.format)
		{
		case FMT_VGM: strncat(outname,".vgm",sizeof(outname)); break;
		case FMT_TXT: strncat(outname,".txt",sizeof(outname)); break;
		}

		convert_song(outname,song);
	}

	free(NSF.data);

	return 0;
}